@@ -22,24 +22,33 @@ SEXP setattrib(SEXP x, SEXP name, SEXP value)
     UNPROTECT(1);
     return(x);
   }
-  setAttrib(x, name,
-    MAYBE_REFERENCED(value) ? duplicate(value) : value);
+  if (isNull(value) && isPairList(x) && strcmp(CHAR(STRING_ELT(name,0)),"names")==0) {
+    // backport fix in R 3.2.0 to support R 3.1.0; #4048 #3802
+    // apply this backport always (i.e. in R >=3.2.0 too) to avoid a switch on version number or feature test (to avoid more code, tests and nocov)
+    for (SEXP t=x; t!=R_NilValue; t=CDR(t)) {
+      SET_TAG(t, R_NilValue);
+    }
+  } else {
+    setAttrib(x, name, MAYBE_REFERENCED(value) ? duplicate(value) : value);
     // duplicate is temp fix to restore R behaviour prior to R-devel change on 10 Jan 2014 (r64724).
     // TO DO: revisit. Enough to reproduce is: DT=data.table(a=1:3); DT[2]; DT[,b:=2]
     // ... Error: selfrefnames is ok but tl names [1] != tl [100]
+  }
   return(R_NilValue);
 }
 
 // fix for #1142 - duplicated levels for factors
 SEXP setlevels(SEXP x, SEXP levels, SEXP ulevels) {
 
-  R_len_t nx = length(x), i;
+  R_len_t nx = length(x);
   SEXP xchar, newx;
   xchar = PROTECT(allocVector(STRSXP, nx));
-  for (i=0; i<nx; i++)
-    SET_STRING_ELT(xchar, i, STRING_ELT(levels, INTEGER(x)[i]-1));
+  int *ix = INTEGER(x);
+  for (int i=0; i<nx; ++i)
+    SET_STRING_ELT(xchar, i, STRING_ELT(levels, ix[i]-1));
   newx = PROTECT(chmatch(xchar, ulevels, NA_INTEGER));
-  for (i=0; i<nx; i++) INTEGER(x)[i] = INTEGER(newx)[i];
+  int *inewx = INTEGER(newx);
+  for (int i=0; i<nx; ++i) ix[i] = inewx[i];
   setAttrib(x, R_LevelsSymbol, ulevels);
   UNPROTECT(2);
   return(x);
@@ -62,14 +71,6 @@ SEXP setlistelt(SEXP l, SEXP i, SEXP value)
   return(R_NilValue);
 }
 
-SEXP setmutable(SEXP x)
-{
-  // called from one single place at R level. TODO: avoid somehow, but fails tests without
-  // At least the SET_NAMED() direct call is passed 0 and makes no assumptions about >0.  Good enough for now as patch for CRAN is needed.
-  SET_NAMED(x,0);
-  return(x);
-}
-
 SEXP address(SEXP x)
 {
   // A better way than : http://stackoverflow.com/a/10913296/403310
@@ -78,32 +79,11 @@ SEXP address(SEXP x)
   return(mkString(buffer));
 }
 
-SEXP copyNamedInList(SEXP x)
-{
-  // As from R 3.1.0 list() no longer copies NAMED inputs
-  // Since data.table allows subassignment by reference, we need a way to copy NAMED inputs, still.
-  // But for many other applications (such as in j and elsewhere internally) the new non-copying list() in R 3.1.0 is very welcome.
-
-  // This is intended to be called just after list(...) in data.table().  It isn't for use on a single data.table, as
-  // member columns of a list aren't marked as NAMED when the VECSXP is.
-
-  // For now, this makes the old behaviour of list() in R<3.1.0 available for use, where we need it.
-
-  if (TYPEOF(x) != VECSXP) error("x isn't a VECSXP");
-  for (int i=0; i<LENGTH(x); i++) {
-    SEXP col = VECTOR_ELT(x, i);
-    if (MAYBE_REFERENCED(col) || ALTREP(col)) {
-      SET_VECTOR_ELT(x, i, duplicate(col));
-    }
-  }
-  return R_NilValue;
-}
-
 SEXP expandAltRep(SEXP x)
 {
   // used by setDT to ensure altrep vectors in columns are expanded. Such altrep objects typically come from tests or demos, since
   // the sequence 1:n does not occur in real-world data as a column, very often.
-  // Note that data.table() calls copyNamedInList() (above) which has the side-effect of expanding altrep vectors too.
+  // Note that data.table() calls as.data.table.list() which expands altrep vectors.
   // We need regular expanded columns in data.table because `:=` relies on that, for example.
   // At R level (for example [.data.table) we use and benefit from altrep vectors very much. It's just as columns that we expand them.
   // See extensive discussion in issue #2866
@@ -112,7 +92,7 @@ SEXP expandAltRep(SEXP x)
   for (int i=0; i<LENGTH(x); i++) {
     SEXP col = VECTOR_ELT(x,i);
     if (ALTREP(col)) {
-      SET_VECTOR_ELT(x, i, duplicate(col));
+      SET_VECTOR_ELT(x, i, copyAsPlain(col));
     }
   }
   return R_NilValue;
